@@ -1,18 +1,24 @@
-from nose.tools import assert_equal
-
+import sys
+import tempfile
 from copy import copy
 from pprint import pformat
-from unittest import TestCase
+from textwrap import dedent
+from StringIO import StringIO
 from xml.dom import Node
 
+from nose.tools import assert_equal, assert_raises
+from unittest import TestCase
+
 from htmltreediff import html_changes
 from htmltreediff.text import text_changes, WordMatcher, PlaceholderMatcher
 from htmltreediff.html import distribute, fix_lists, fix_tables
+from htmltreediff.cli import main
 from htmltreediff.util import (
     parse_minidom,
     minidom_tostring,
     html_equal,
     get_location,
+    remove_insignificant_text_nodes,
 )
 from htmltreediff.test_util import (
     reverse_edit_script,
@@ -31,8 +37,142 @@ def test_cutoff():
         '<h1>totally</h1>',
         '<h2>different</h2>',
     )
-    assert changes == ('<h2>The differences from the previous version are too '
-                       'large to show concisely.</h2>')
+    assert_equal(
+        changes,
+        '<h2>The differences from the previous version are too large to show '
+        'concisely.</h2>',
+    )
+
+def test_illegal_text_nodes():
+    html = '''
+        <table>
+            illegal text
+            <tr>
+                <td>stuff</td>
+            </tr>
+        </table>
+    '''
+    dom = parse_minidom(html)
+    html = minidom_tostring(dom)
+    assert_equal(
+        html,
+        '<html><head/><body> illegal text '
+        '<table><tbody><tr><td>stuff</td></tr></tbody></table></body></html>',
+    )
+
+def test_remove_insignificant_text_nodes():
+    html = dedent('''
+        <html>
+            <head />
+            <body>
+                <p>
+                    one <em>two</em> <strong>three</strong>
+                </p>
+                <table>
+                    <tr>
+                        <td>stuff</td>
+                    </tr>
+                </table>
+            </body>
+        </html>
+    ''')
+    dom = parse_minidom(html)
+    remove_insignificant_text_nodes(dom)
+    html = minidom_tostring(dom)
+    assert_equal(
+        html,
+        '<html><head/><body> <p> one <em>two</em> <strong>three</strong> </p> '
+        '<table><tbody><tr><td>stuff</td></tr></tbody></table> </body></html>',
+    )
+
+    # Check that it is idempotent.
+    dom = parse_minidom(html)
+    remove_insignificant_text_nodes(dom)
+    html = minidom_tostring(dom)
+    assert_equal(
+        html,
+        ('<html><head/><body> <p> one <em>two</em> <strong>three</strong> </p> '
+         '<table><tbody><tr><td>stuff</td></tr></tbody></table> </body></html>'),
+    )
+
+def test_remove_insignificant_text_nodes_nbsp():
+    html = dedent('''
+        <table>
+        <tbody>
+        <tr>
+            <td> </td>
+            <td>&#160;</td>
+            <td>&nbsp;</td>
+        </tr>
+        </tbody>
+        </table>
+    ''')
+    dom = parse_minidom(html)
+    remove_insignificant_text_nodes(dom)
+    html = minidom_tostring(dom)
+    assert_equal(
+        html,
+        ('<html><head/><body><table><tbody><tr><td> </td><td> </td><td> </td>'
+         '</tr></tbody></table></body></html>'),
+    )
+
+def test_html_changes_pretty():
+    cases = [
+        (
+            'Simple Addition',
+            '<h1>one</h1>',
+            '<h1>one</h1><h2>two</h2>',
+            dedent('''
+                <h1>
+                  one
+                </h1>
+                <ins>
+                  <h2>
+                    two
+                  </h2>
+                </ins>
+            ''').strip(),
+        ),
+    ]
+    for test_name, old_html, new_html, pretty_changes in cases:
+        def test():
+            changes = html_changes(old_html, new_html, cutoff=0.0, pretty=True)
+            print changes
+            assert_equal(pretty_changes, changes)
+        test.description = 'test_html_changes_pretty - %s' % test_name
+        yield test
+
+def test_main():
+    # Run the command line interface main function.
+    f1 = tempfile.NamedTemporaryFile()
+    f1.write('<h1>one</h1>')
+    f1.seek(0)
+    f2 = tempfile.NamedTemporaryFile()
+    f2.write('<h1>one</h1><h2>two</h2>')
+    f2.seek(0)
+
+    try:
+        old_stdout = sys.stdout
+        sys.stdout = stream = StringIO()
+        main(argv=('', f1.name, f2.name))
+        assert_equal(
+            stream.getvalue(),
+            dedent('''
+                <h1>
+                  one
+                </h1>
+                <ins>
+                  <h2>
+                    two
+                  </h2>
+                </ins>
+            ''').strip() + '\n',
+        )
+    finally:
+        sys.stdout = old_stdout
+
+    # Run it with no arguments, it throws an error.
+    assert_raises(IOError, main)
 
 
 class TextChangesTestCase(TestCase):
@@ -660,7 +800,8 @@ insane_test_cases = [
     ( # delete top row and add a column, funny whitespace
         '<table> <tr><td>A1</td></tr> <tr><td>B1</td></tr> </table>',
         '<table> <tr><td>B1</td><td>B2</td></tr> </table>',
-        '<table> <tr><td><del>A1</del></td></tr> <tr><td>B1</td><td><ins>B2</ins></td></tr> </table>',
+        ('<table><tbody><tr><td><del>A1</del></td></tr><tr><td>B1</td>'
+         '<td><ins>B2</ins></td></tr></tbody></table>'),
         [],
     ),
     ( # handle newline-separated words correctly
@@ -719,6 +860,42 @@ insane_test_cases = [
         '''),
         [],
     ),
+    ( # whitespace changes in a table with nbsp entity
+        '''
+        <table>
+        <tbody>
+        <tr>
+            <td> </td>
+            <td>&#160;</td>
+            <td>&nbsp;</td>
+        </tr>
+        </tbody>
+        </table>
+        ''',
+        '''
+        <table>
+        <tbody>
+        <tr>
+        <td> </td>
+        <td>&#160;</td>
+        <td>&nbsp;</td>
+        </tr>
+        </tbody>
+        </table>
+        ''',
+        collapse('''
+        <table>
+        <tbody>
+        <tr>
+        <td> </td>
+        <td>&#160;</td>
+        <td>&nbsp;</td>
+        </tr>
+        </tbody>
+        </table>
+        '''),
+        [],
+    ),
 #BROKEN, see issue #2384
 #    # ul and ol tags are considered equal when diffing
 #    (
@@ -1092,6 +1269,37 @@ class HtmlChangesTestCase(TestCase):
                 </table>
                 '''
             ),
+            ( # remove ins and del tags at the wrong level of the table
+                '''
+                <table>
+                    <ins> </ins><del> </del>
+                    <thead>
+                        <ins> </ins><del> </del>
+                    </thead>
+                    <tfoot>
+                        <ins> </ins><del> </del>
+                    </tfoot>
+                    <tbody>
+                        <ins> </ins><del> </del>
+                        <tr>
+                            <ins> </ins><del> </del>
+                            <td><ins>A</ins></td>
+                        </tr>
+                    </tbody>
+                </table>
+                ''',
+                '''
+                <table>
+                    <thead></thead>
+                    <tfoot></tfoot>
+                    <tbody>
+                        <tr>
+                            <td><ins>A</ins></td>
+                        </tr>
+                    </tbody>
+                </table>
+                ''',
+            ),
         ]
         for changes, fixed_changes in cases:
             changes = collapse(changes)
